import json
import os
import pickle
import socket
from dataclasses import dataclass, replace


def load(file_path: str):
    with open(file_path, "rb") as data:
        return pickle.load(data)


def from_json(file_path: str):
    with open(file_path, "r") as file:
        loaded = json.load(file)
        return Parameters(
            host=loaded.get('host') or "localhost",
            client_port=loaded.get('client_port') or 5222,
            server_port=loaded.get('server_port') or 5269,
            server_out_port=loaded.get('server_out_port') or 5269,
            family=loaded.get('family') or socket.AF_INET,
            connection_timeout=loaded.get('connection_timeout') or 60,
            database_path=loaded.get('database_path') or os.path.join(os.getcwd(), "pyjabber.db"),
            database_purge=loaded.get('database_purge') or False,
            database_in_memory=loaded.get('database_in_memory') or False,
            cert_path=loaded.get('cert_path') or None,
            message_persistence=loaded.get('message_persistence') is True
        )


@dataclass(frozen=True)
class Parameters:
    host: str = "localhost"
    client_port: int = 5222
    server_port: int = 5269
    server_out_port: int = 5269
    family: socket.AddressFamily = socket.AF_INET
    connection_timeout: int = 60
    database_path: str = os.path.join(os.getcwd(), "pyjabber.db")
    database_purge: bool = False
    database_in_memory: bool = False
    cert_path: str = None
    message_persistence: bool = True

    def dump(self, file_path: str):
        with open(file_path, "wb") as data:
            pickle.dump(self, data)

    def update_from_json(self, file_path: str):
        with open(file_path, "r") as file:
            loaded = json.load(file)
            return replace(
                self,
                host=loaded.get('host') or "localhost",
                client_port=loaded.get('client_port') or 5222,
                server_port=loaded.get('server_port') or 5269,
                server_out_port=loaded.get('server_out_port') or 5269,
                family=loaded.get('family') or socket.AF_INET,
                connection_timeout=loaded.get('connection_timeout') or 60,
                database_path=loaded.get('database_path') or os.path.join(os.getcwd(), "pyjabber.db"),
                database_purge=loaded.get('database_purge') or False,
                database_in_memory=loaded.get('database_in_memory') or False,
                cert_path=loaded.get('cert_path') or None,
                message_persistence=loaded.get('message_persistence') is True
            )



import json
import os
import socket
from unittest.mock import mock_open, patch

from pyjabber.server_parameters import Parameters, from_json, load


def test_default_parameters():
    param = Parameters()
    assert param.host == "localhost"
    assert param.client_port == 5222
    assert param.server_port == 5269
    assert param.server_out_port == 5269
    assert param.family == socket.AF_INET
    assert param.connection_timeout == 60
    assert param.database_path == os.path.join(os.getcwd(), "pyjabber.db")
    assert param.database_purge is False
    assert param.database_in_memory is False
    assert param.cert_path is None
    assert param.message_persistence is True

def test_from_json():
    datamock = {
        'host': 'publichost',
        'client_port': 5252,
        'server_port': 2525,
        'server_out_port': 6952,
        'family': socket.AF_INET6.value,
        'connection_timeout': 59,
        'database_path': '/some/fake/path',
        'database_purge': True,
        'database_in_memory': True,
        'cert_path': '/another/fake/path',
        'message_persistence': False
    }
    m = mock_open(read_data=json.dumps(datamock))

    with patch("pyjabber.server_parameters.open", m):
        param = from_json("fake.json")

    assert param.host == datamock["host"]
    assert param.client_port == datamock["client_port"]
    assert param.server_port == datamock["server_port"]
    assert param.server_out_port == datamock["server_out_port"]
    assert param.family == datamock["family"]
    assert param.connection_timeout == datamock["connection_timeout"]
    assert param.database_path == datamock["database_path"]
    assert param.database_purge is datamock["database_purge"]
    assert param.database_in_memory is datamock["database_in_memory"]
    assert param.cert_path == datamock["cert_path"]
    assert param.message_persistence is datamock["message_persistence"]

def test_update_from_json():
    datamock = {
        'host': 'publichost',
        'client_port':  5252,
        'server_port': 2525,
        'server_out_port': 6952,
        'family': socket.AF_INET6.value,
        'connection_timeout': 59,
        'database_path': '/some/fake/path',
        'database_purge': True,
        'database_in_memory': True,
        'cert_path': '/another/fake/path',
        'message_persistence': False
    }
    m = mock_open(read_data=json.dumps(datamock))

    with patch("pyjabber.server_parameters.open", m):
        param = Parameters().update_from_json("fake.json")

    assert param.host == datamock["host"]
    assert param.client_port == datamock["client_port"]
    assert param.server_port == datamock["server_port"]
    assert param.server_out_port == datamock["server_out_port"]
    assert param.family == datamock["family"]
    assert param.connection_timeout == datamock["connection_timeout"]
    assert param.database_path == datamock["database_path"]
    assert param.database_purge is datamock["database_purge"]
    assert param.database_in_memory is datamock["database_in_memory"]
    assert param.cert_path == datamock["cert_path"]
    assert param.message_persistence is datamock["message_persistence"]


def test_pickle():
    filepath = './params.pkl'
    param = Parameters()
    param.dump(filepath)

    assert os.path.exists(filepath)

    parsed_param = load(filepath)

    assert parsed_param.host == "localhost"
    assert parsed_param.client_port == 5222
    assert parsed_param.server_port == 5269
    assert parsed_param.server_out_port == 5269
    assert parsed_param.family == socket.AF_INET
    assert parsed_param.connection_timeout == 60
    assert parsed_param.database_path == os.path.join(os.getcwd(), "pyjabber.db")
    assert parsed_param.database_purge is False
    assert parsed_param.database_in_memory is False
    assert parsed_param.cert_path is None
    assert parsed_param.message_persistence is True

    if os.path.exists(filepath):
        os.remove(filepath)



import os
import sqlite3
from contextlib import closing
from typing import Optional

from pyjabber import __version__
from pyjabber import metadata
from loguru import logger


def connection() -> sqlite3.Connection:
    """
    Returns an already crafted connection with the database.
    It takes the parameters from the server class instance (i.e., DB path | DB in memory)
    """
    if metadata.database_in_memory.get() is not None:
        return sqlite3.connect("file::memory:?cache=shared", uri=True)
    return sqlite3.connect(metadata.database_path.get())


def setup_database(
    sql_init_script: str,
    database_in_memory: bool = False,
    database_path: str = None,
    database_purge: bool = False
):
    if database_in_memory:
        logger.info("Using database on memory. ANY CHANGE WILL BE LOST AFTER SERVER SHUTDOWN!")
        db_in_memory_con = sqlite3.connect("file::memory:?cache=shared", uri=True)
        with open(sql_init_script, "r") as script_template:
            script = script_template.read()
            script = script.replace("{{version}}", f"'{__version__}'")
            db_in_memory_con.cursor().executescript(script)
            db_in_memory_con.commit()
        metadata.database_in_memory.set(db_in_memory_con)

    elif os.path.isfile(database_path) is False:
        logger.info("No database found. Initializing one...")
        if database_purge:
            logger.info("Ignoring purge database flag. No DB to purge")
        with closing(connection()) as con:
            with open(sql_init_script, "r") as script_template:
                script = script_template.read()
                script = script.replace("{{version}}", f"'{__version__}'")
                con.cursor().executescript(script)
            con.commit()
    else:
        if database_purge:
            logger.info("Resetting the database to default state...")
            os.remove(database_path)
            with closing(connection()) as con:
                with open(sql_init_script, "r") as script_template:
                    script = script_template.read()
                    script = script.replace("{{version}}", f"'{__version__}'")
                    con.cursor().executescript(script)
                con.commit()


def migration(version: Optional[str]): # pragma: no cover
    if not version: # pragma: no cover
        return # pragma: no cover

    # DB migration feature was added in v0.2.6
    # Only changes in the db after v0.2.6 will be taken into account for migration




import os
import socket
import sqlite3
from contextlib import closing

from loguru import logger

import pyjabber
from . import metadata
from pyjabber.db.database import connection
from pyjabber.network import CertGenerator


def setup_query_local_ip():
    """Return the local IP of the host machine"""
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


def setup_ip_by_host(host: str):
    try:
        return socket.gethostbyname(host)
    except socket.gaierror as e:
        logger.error(e)
        return None


def setup_certs(host: str, cert_path: str):
    try:
        if CertGenerator.check_hostname_cert_exists(host, cert_path) is False:
            CertGenerator.generate_hostname_cert(host, cert_path)
    except FileNotFoundError as e:
        logger.error(f"{e.__class__.__name__}: Pass an existing directory in your system to load the certs. "
                     f"Closing server")
        raise SystemExit
